import json
import os
import shutil

import httpx
from httpx import Response
from jinja2 import Template

from hotbox.const import (
    DEFAULT_IMAGE_TEMPLATE_DIR,
    DEFAULT_LANG_TEMPLATE_DIR,
    DEFAULT_RUN_APP_TEMPLATE_FILEPATH,
)
from hotbox.settings import env
from hotbox.types import Image, Routes


class AppService:
    def __init__(self) -> None:
        pass

    def create_app_bundle(
        self,
        app_id: str,
        app_code_path: str,
        build_image: Image,
        vcpu_count: int,
        mem_size_mib: int,
        tmpdir: str,
    ) -> str:
        _image_dir = f"{tmpdir}/{app_id}_image"
        _code_dir = f"{tmpdir}/{app_id}_code"
        os.makedirs(_image_dir, exist_ok=True)
        os.makedirs(_code_dir, exist_ok=True)
        shutil.copytree(
            src=f"{DEFAULT_IMAGE_TEMPLATE_DIR}",
            dst=_image_dir,
            dirs_exist_ok=True,
        )
        shutil.copytree(
            src=app_code_path,
            dst=_code_dir,
            dirs_exist_ok=True,
        )
        self._create_image(
            image=build_image,
            image_dir=_image_dir,
        )
        self._create_run_app(
            app_id=app_id,
            vcpu_count=vcpu_count,
            mem_size_mib=mem_size_mib,
            tmpdir=tmpdir,
        )
        # tar up the bundle
        bundle_path = f"{tmpdir}/{app_id}"
        shutil.make_archive(
            base_name=bundle_path,
            format="gztar",
            root_dir=tmpdir,
        )
        return f"{bundle_path}.tar.gz"

    def _create_image(self, image: Image, image_dir: str) -> None:
        self._create_dockerfile(
            image_dir=image_dir,
            image=image,
        )
        self._create_entrypoint(
            image_dir=image_dir,
            image=image,
        )
        self._create_start_script(
            image_dir=image_dir,
            image=image,
        )

    def _create_dockerfile(
        self,
        image_dir: str,
        image: Image,
    ) -> None:
        with open(f"{image_dir}/Dockerfile.j2") as f:
            template = Template(f.read()).render(
                image=image.value,
            )
        with open(f"{image_dir}/Dockerfile", "w") as f:
            f.write(template)
        os.remove(f"{image_dir}/Dockerfile.j2")

    def _create_entrypoint(self, image_dir: str, image: Image) -> None:
        _install = self._get_install_steps(image=image)
        _build = self._get_build_steps(image=image)
        with open(f"{image_dir}/entrypoint.j2") as f:
            template = Template(f.read()).render(
                install=_install,
                build=_build,
            )
        with open(f"{image_dir}/entrypoint", "w") as f:
            f.write(template)
        os.remove(f"{image_dir}/entrypoint.j2")

    def _get_install_steps(self, image: Image) -> str:
        with open(f"{DEFAULT_LANG_TEMPLATE_DIR}/{image.name}/install") as f:
            return f.read().strip()

    def _get_build_steps(self, image: Image) -> str:
        with open(f"{DEFAULT_LANG_TEMPLATE_DIR}/{image.name}/build") as f:
            return f.read().strip()

    def _create_start_script(self, image_dir: str, image: Image) -> None:
        _entrypoint = self._get_entrypoint(image=image)
        with open(f"{image_dir}/start.sh.j2") as f:
            template = Template(f.read()).render(
                image=image,
                entrypoint=_entrypoint,
            )
        with open(f"{image_dir}/start.sh", "w") as f:
            f.write(template)
        os.remove(f"{image_dir}/start.sh.j2")

    def _get_entrypoint(self, image: Image) -> str:
        with open(f"{DEFAULT_LANG_TEMPLATE_DIR}/{image.name}/entrypoint") as f:
            return f.read().strip()

    def _create_run_app(
        self, app_id: str, vcpu_count: int, mem_size_mib: int, tmpdir: str
    ) -> None:
        with open(DEFAULT_RUN_APP_TEMPLATE_FILEPATH) as f:
            template = Template(f.read()).render(
                app_id=app_id,
                vcpu_count=vcpu_count,
                mem_size_mib=mem_size_mib,
            )
        with open(f"{tmpdir}/{app_id}_run_app.sh", "w") as f:
            f.write(template)

    def upload_app_bundle(self, app_id: str, bundle_path: str) -> Response:
        response = httpx.post(
            url=env.HOTBOX_API_URL + Routes.create_apps,
            files={
                "upload_file": (
                    os.path.basename(bundle_path),
                    open(bundle_path, "rb"),
                    "application/gzip",
                ),
                "create_app_request": (
                    None,
                    json.dumps({"app_id": app_id}),
                    "application/json",
                ),
            },
        )
        return response

    def unzip_and_run(self, bundle_path: str, app_id: str) -> None:  # pragma: no cover
        os.system(f"tar -xzf {bundle_path}")
        os.system(f"chmod +x {app_id}_run_app.sh")
        os.system(f"./{app_id}_run_app.sh &")
        os.remove(bundle_path)


app_svc = AppService()



import tempfile
from typing import List

import orjson
from typer import Exit, FileText, Option, Typer, echo

from hotbox.app import app_svc
from hotbox.ec2 import ec2_svc
from hotbox.types import Ec2Spec, Image, Language
from hotbox.utils import determine_lang, generate_app_id, handle_filetext

app = Typer(
    name="create",
    help="Create resources.",
    no_args_is_help=True,
)


@app.command(
    "ec2",
    help="Create ec2 resources.",
    no_args_is_help=True,
)
def create_ec2(
    firecracker_version: str = Option(
        "v1.3.1",
        "-fcv",
        "--firecracker-version",
        help="Firecracker version.",
    ),
    region: str = Option(
        None,
        "-r",
        "--region",
        help="AWS region.",
    ),
    key_name: str = Option(
        None,
        "-k",
        "--key-name",
        help="AWS key name. Should be in the region you specify.",
    ),
    security_group_ids: List[str] = Option(
        None,
        "-s",
        "--security-group-ids",
        help="AWS security group ids. Should be in the region you specify.",
    ),
    _filetext: FileText = Option(
        None,
        "-f",
        "--file",
        allow_dash=True,
        help="Path to the spec file. Accepts JSON strings as input too."
        " Overrides all other options.",
    ),
) -> None:
    """Create ec2 resources.

    This command creates ec2 resources. You can specify certain args directly as listed
    below, or you can specify a spec file with the `-f` or `--file` option. If you
    specify a spec file, all other options will be ignored.

    Args:
        firecracker_version (str, optional): Firecracker version. Defaults to "v1.3.1".
        region (str, optional): AWS region. Defaults to "us-east-1".
        key_name (str, optional): AWS key name. Should be in the region you specify.
            Defaults to None.
        security_group_ids (List[str], optional): AWS security group ids. Should be in
            the region you specify. Defaults to None.
        _filetext (FileText, optional): Path to the spec file. Accepts JSON strings as
            input too. Overrides all other options. Defaults to None.
    """
    echo("Creating ec2!")
    if _filetext is not None:
        content = Ec2Spec(**handle_filetext(str(_filetext)))
    else:
        content = Ec2Spec(
            region=region,
            key_name=key_name,
            security_group_ids=security_group_ids,
        )
    response = ec2_svc.create(
        spec=content,
        firecracker_version=firecracker_version,
    )
    echo(orjson.dumps(response))


@app.command(
    "app",
    help="Deploy apps.",
    no_args_is_help=True,
)
def create_app(
    app_code_path: str = Option(
        ...,
        "-c",
        "--code",
        help="Path to the app code.",
    ),
    vcpu_count: int = Option(
        1,
        "-v",
        "--vcpu-count",
        help="Number of vcpus.",
    ),
    mem_size_mib: int = Option(
        256,
        "-m",
        "--mem-size-mib",
        help="Memory size in MiB.",
    ),
) -> None:
    echo("Creating app!")
    app_id = generate_app_id()
    lang = determine_lang(app_code_path=app_code_path)
    if lang is None:
        echo(
            message="Unsupported application content. "
            + "Please use a supported language: "
            + ",".join(list(map(lambda x: x.name, Language))),
            err=True,
        )
        raise Exit(1)
    build_image = getattr(Image, lang.name)
    with tempfile.TemporaryDirectory() as tmpdir:
        bundle_path = app_svc.create_app_bundle(
            app_id=app_id,
            app_code_path=app_code_path,
            build_image=build_image,
            vcpu_count=vcpu_count,
            mem_size_mib=mem_size_mib,
            tmpdir=tmpdir,
        )
        response = app_svc.upload_app_bundle(
            app_id=app_id,
            bundle_path=bundle_path,
        )
    if not response.is_success:
        echo(
            message=(
                "Failed to upload app bundle with status code "
                f"{response.status_code}, and message: "
                f"{response.json().get('message')}"
            ),
            err=True,
        )
        raise Exit(1)
    echo(orjson.dumps(response.json()))



import uvicorn
from typer import Option, Typer

from hotbox.const import SERVER_MODULE_NAME

app = Typer(
    name="server",
    help="Run the hotbox server.",
    no_args_is_help=True,
)


@app.command(
    "run",
    help="Run the hotbox server.",
)
def server(
    host: str = Option(
        "127.0.0.1",
        "--host",
        help="Bind socket to this host.",
    ),
    port: int = Option(
        8420,
        "--port",
        help="Bind socket to this port.",
    ),
    reload: bool = Option(
        False,
        help="Reload the server on code changes.",
    ),
    workers: int = Option(
        1,
        "--workers",
        help="Number of workers to run.",
    ),
) -> None:
    uvicorn.run(  # pragma: no cover
        app=SERVER_MODULE_NAME,
        port=port,
        host=host,
        reload=reload,
        workers=workers,
    )



import os
import sys

from pydantic import BaseSettings, Field


class _Env(BaseSettings):
    HOTBOX_API_URL: str = Field(
        "http://localhost:8420/api/v0",
        env="HOTBOX_API_URL",
        description="hotbox API URL",
    )

    class Config:
        env_file = ".env.local"
        env_file_encoding = "utf-8"


if os.getenv("_", "").endswith("pytest") or "pytest" in "".join(sys.argv):
    _Env.Config.env_file = ".env.test"  # type: ignore

env = _Env()

# if the ENV_FILE environment variable is set, use it as an override
# this is useful for running alembic migrations against remote databases
if os.getenv("ENV_FILE") is not None:
    env = _Env(_env_file=os.environ["ENV_FILE"])  # pragma: no cover



import os
from datetime import datetime

from fastapi import APIRouter, BackgroundTasks, FastAPI, File, Form, UploadFile
from fastapi.responses import JSONResponse
from pydantic import Json

from hotbox import __version__
from hotbox.app import app_svc
from hotbox.const import API_V0, DESC, NAME
from hotbox.types import CreateAppRequest, HealthcheckResponse

os.environ["TZ"] = "UTC"

api = FastAPI(
    title=NAME,
    desc=DESC,
    version=__version__,
)

health_router = APIRouter(
    prefix=API_V0,
    tags=["health"],
)

app_router = APIRouter(
    prefix=API_V0,
    tags=["app"],
)


@health_router.get(
    "/healthcheck",
    response_model=HealthcheckResponse,
)
async def healthcheck() -> HealthcheckResponse:
    return HealthcheckResponse(
        message="ok",
        version=__version__,
        time=datetime.utcnow(),
    )


@app_router.post(
    "/apps",
    response_class=JSONResponse,
)
async def create_app(
    background_tasks: BackgroundTasks,
    create_app_request: Json = Form(...),
    upload_file: UploadFile = File(...),
) -> JSONResponse:
    _create_app_request = CreateAppRequest(**create_app_request)
    bundle_path = f"{_create_app_request.app_id}.tar.gz"
    with open(bundle_path, "wb") as buffer:
        buffer.write(upload_file.file.read())
    background_tasks.add_task(
        app_svc.unzip_and_run,
        bundle_path,
        _create_app_request.app_id,
    )
    return JSONResponse(
        status_code=200,
        content={"message": "App running in the cloud!"},
    )


api.include_router(health_router)
api.include_router(app_router)



import orjson
from typer import Option, Typer, echo

from hotbox.ec2 import ec2_svc

app = Typer(
    name="get",
    help="Get resources.",
    no_args_is_help=True,
)


@app.command(
    "ec2",
    help="Get ec2 resources.",
    no_args_is_help=True,
)
def get_ec2(
    region: str = Option(
        None,
        help="AWS Region.",
    ),
) -> None:
    response = ec2_svc.get(
        region=region,
    )
    echo(orjson.dumps(response))



from typing import List

import orjson
from typer import Argument, Option, Typer, echo

from hotbox.ec2 import ec2_svc

app = Typer(
    name="delete",
    help="Delete resources.",
    no_args_is_help=True,
)


@app.command(
    "ec2",
    help="Delete ec2 resources.",
    no_args_is_help=True,
)
def delete_ec2(
    ids: List[str] = Argument(
        ...,
        help="IDs of resources to delete.",
    ),
    region: str = Option(
        None,
        help="AWS Region.",
    ),
) -> None:
    response = ec2_svc.delete(ids, region)
    echo(orjson.dumps(response))



from datetime import datetime
from enum import Enum, unique
from typing import Any, Dict, List

from pydantic import BaseModel, StrictStr


@unique
class Language(str, Enum):
    go = "go"
    # python = "py"


@unique
class Image(str, Enum):
    """Image is the name of the Docker image to use for hotbox."""

    go = "golang:1.20"
    # python = "python:3.11-slim"


class Routes(str, Enum):
    create_apps = "/apps"


@unique
class Ec2MetalType(str, Enum):
    """Ec2MetalType is the type of EC2 instance to use for hotbox."""

    a1_metal = "a1.metal"
    i3_metal = "i3.metal"
    g4dn_metal = "g4dn.metal"


class HotboxSpec(BaseModel):
    ...


class AwsSpec(HotboxSpec):
    region: str


class Ec2Spec(AwsSpec):
    key_name: str
    security_group_ids: List[str]
    instance_type: Ec2MetalType = Ec2MetalType.a1_metal
    min_count: int = 1
    max_count: int = 1
    monitoring_enabled: Dict[str, Any] = {
        "Enabled": False,
    }
    metadata_options: Dict[str, Any] = {
        "InstanceMetadataTags": "enabled",
    }
    block_device_mappings: List[Dict[str, Any]] = [
        {
            "DeviceName": "/dev/xvda",
            "Ebs": {
                "DeleteOnTermination": True,
                "VolumeSize": 8,
                "VolumeType": "gp2",
            },
        },
    ]


class HealthcheckResponse(BaseModel):
    message: StrictStr
    version: StrictStr
    time: datetime


class CreateAppRequest(BaseModel):
    app_id: StrictStr



import json
import os
import sys
import uuid
from typing import Dict, Optional

from hotbox.types import Language


def handle_filetext(filetext: str) -> Dict:
    if os.path.isfile(filetext):
        with open(filetext, "r") as f:
            content = f.read().encode("utf-8")
    elif filetext == "-":
        content = sys.stdin.buffer.read()
    else:
        content = filetext.encode("utf-8")
    return json.loads(content)


def generate_app_id() -> str:
    return f"app-{uuid.uuid4().hex}"


def determine_lang(app_code_path: str) -> Optional[Language]:
    for _, _, filenames in os.walk(app_code_path):
        for filename in filenames:
            lang = os.path.splitext(filename)[-1].split(".")[-1]
            if lang in list(Language):
                return Language(lang)
    return None



from typing import Dict, List, Optional

import boto3
from jinja2 import Template

from hotbox.const import DEFAULT_USERDATA_TEMPLATE_FILEPATH
from hotbox.types import Ec2Spec


class Ec2Service:
    def ec2_client(self, region_name: str) -> boto3.client:
        return boto3.client("ec2", region_name=region_name)  # pragma: no cover

    def get_image_id(
        self,
        ec2_client: boto3.client,
        name: str = "ubuntu/images/hvm-ssd/ubuntu-bionic-18.04-arm64-*",
        virtualization_type: str = "hvm",
        architecture: str = "arm64",
        root_device_type: str = "ebs",
        owner_id: str = "099720109477",
    ) -> str:
        filters = [
            {"Name": "name", "Values": [name]},
            {"Name": "virtualization-type", "Values": [virtualization_type]},
            {"Name": "architecture", "Values": [architecture]},
            {"Name": "root-device-type", "Values": [root_device_type]},
            {"Name": "owner-id", "Values": [owner_id]},
        ]
        images = ec2_client.describe_images(Filters=filters)
        images_sorted = sorted(
            images["Images"],
            key=lambda x: x["CreationDate"],
            reverse=True,
        )
        return images_sorted[0]["ImageId"]

    def _template_userdata(
        self,
        firecracker_version: str,
        userdata_template_filepath: Optional[str] = None,
    ) -> str:
        if userdata_template_filepath is None:
            userdata_template_filepath = DEFAULT_USERDATA_TEMPLATE_FILEPATH
        with open(userdata_template_filepath) as f:
            return Template(f.read()).render(firecracker_version=firecracker_version)

    def create(self, spec: Ec2Spec, firecracker_version: str) -> Dict:
        ec2_client = self.ec2_client(region_name=spec.region)
        return ec2_client.run_instances(
            ImageId=self.get_image_id(ec2_client=ec2_client),
            KeyName=spec.key_name,
            InstanceType=spec.instance_type,
            MinCount=spec.min_count,
            MaxCount=spec.max_count,
            Monitoring=spec.monitoring_enabled,
            SecurityGroupIds=spec.security_group_ids,
            BlockDeviceMappings=spec.block_device_mappings,
            MetadataOptions=spec.metadata_options,
            UserData=self._template_userdata(firecracker_version=firecracker_version),
        )

    def get(self, region: str) -> Dict:
        ec2_client = self.ec2_client(region_name=region)
        return ec2_client.describe_instances()

    def delete(self, ids: List[str], region: str) -> Dict:
        ec2_client = self.ec2_client(region_name=region)
        return ec2_client.terminate_instances(InstanceIds=ids)


ec2_svc = Ec2Service()



from typer import Typer, echo

from hotbox import __version__
from hotbox.cli import create, delete, get, server
from hotbox.const import DESC, NAME

app = Typer(
    name=NAME,
    help=DESC,
    no_args_is_help=True,
)


@app.callback()
def main_callback() -> None:
    pass


main_callback.__doc__ = NAME


@app.command("version")
def _version() -> None:
    """Prints the version."""
    echo(__version__, nl=False)


app.add_typer(create.app)
app.add_typer(get.app)
app.add_typer(delete.app)
app.add_typer(server.app)
